/**
 * @file image.c
 * @brief Image managment
 *
 * @section License
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 *
 * Copyright (C) 2010-2022 Oryx Embedded SARL. All rights reserved.
 *
 * This file is part of CycloneBOOT Open.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * @author Oryx Embedded SARL (www.oryx-embedded.com)
 * @version 2.0.2
 **/

//Dependencies
#include <stdlib.h>
#include "image.h"
#include "core/crc32.h"
#include "debug.h"
#include "core/crypto.h"
#include "core/slot.h"


/**
 * @brief Check Internal Image header validity
 * @param[in] header Pointeur to the internal image header to be checked
 * @return Status code
 **/
 
error_t imageCheckHeader(ImageHeader *header)
{
   error_t error;
   uint32_t computedCrc;
   
   //Check parameter vailidty
   if(header == NULL)
      return ERROR_INVALID_PARAMETER;
   
   //Compute image header crc
   error = CRC32_HASH_ALGO->compute((uint8_t*)header, sizeof(ImageHeader) - CRC32_DIGEST_SIZE, (uint8_t*)&computedCrc);
   if(error)
   {
      //Debug message
      TRACE_ERROR("Failed to compute image header crc!\r\n");
      return error;
   }
   
   //Check image header integrity
   if(header->headCrc != computedCrc)
      return ERROR_INVALID_HEADER;
   
   //Check image header version
   if(header->headVers != IMAGE_HEADER_VERSION)
      return ERROR_INVALID_VERSION;
   
   //Successful process
   return NO_ERROR;
}


/**
 * @brief Get image header from data buffer
 * @param[in] buffer Data buffer containing the image header
 * @param[in] bufferLen Data buffer length
 * @param[out] header Pointeur to the header structure to be returned
 * @return Status code
 **/

error_t imageGetHeader(uint8_t *buffer, size_t bufferLen, ImageHeader **header)
{
   error_t error;
   ImageHeader *tempHeader;
   
   //Check parameters validity
   if(buffer == NULL || bufferLen == 0)
      return ERROR_INVALID_PARAMETER;
   
   if(bufferLen < sizeof(ImageHeader))
      return ERROR_INVALID_LENGTH;
   
   //Point to the image header
   tempHeader = (ImageHeader *)buffer;
   
   //Check image header
   error = imageCheckHeader(tempHeader);
   //Is any error?
   if(error)
      return error;
   
   //Save image header
   *header = (ImageHeader*) tempHeader;
   
   //Successful process
   return NO_ERROR;
}
