/**
 * @file iap_sign.c
 * @brief CycloneBOOT IAP image signature handling
 *
 * @section License
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 *
 * Copyright (C) 2010-2022 Oryx Embedded SARL. All rights reserved.
 *
 * This file is part of CycloneBOOT Open.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * @author Oryx Embedded SARL (www.oryx-embedded.com)
 * @version 2.0.2
 **/

//Switch to the appropriate trace level
#define TRACE_LEVEL IAP_TRACE_LEVEL

//Dependencies
#include "iap/iap_sign.h"
#if IAP_RSA_SUPPORT == ENABLED
#include "pkc/rsa.h"
#endif
#if IAP_ECDSA_SUPPORT == ENABLED
#include "ecc/ecdsa.h"
#endif
#include "pkix/pem_import.h"
#include "debug.h"


//IAP signature private related functions
error_t iapSignVerifyRsa(ImageContext *context);
error_t iapSignVerifyEcdsa(ImageContext *context);


///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


/**
 * @brief Initialize signature material for further signature hash computation
 * and signature verification.
 * @param[in,out] context Pointer to the IAP context
 * @return Error code
 **/

error_t iapSignInit(ImageContext *context)
{
#if IAP_SIGNATURE_SUPPORT == ENABLED
   error_t error;
   IapCryptoSettings* cryptoSettings;

#if IAP_RSA_SUPPORT == ENABLED
   RsaPublicKey publicKey;
#endif
#if IAP_ECDSA_SUPPORT == ENABLED
   EcDomainParameters ecParams;
#endif
   
   //Check parameter validity
   if (context == NULL)
      return ERROR_INVALID_PARAMETER;

   //Point to the IAP user crypto settings
   cryptoSettings = (IapCryptoSettings*)context->imageCryptoSettings;

   //Initialize status code
   error = NO_ERROR;

   //Check user settings
   if (cryptoSettings->signAlgo == IAP_SIGN_NONE || cryptoSettings->signHashAlgo == NULL ||
      cryptoSettings->signKey == NULL || cryptoSettings->signKeyLen == 0)
      return ERROR_INVALID_VALUE;
   
#if IAP_RSA_SUPPORT == ENABLED
   //Is signature RSA algorithm?
   if (cryptoSettings->signAlgo == IAP_SIGN_RSA)
   {
      //Initialize signature algo context
      cryptoSettings->signHashAlgo->init(context->checkContext);

      //Set digest length
      context->digestLen = cryptoSettings->signHashAlgo->digestSize;

      //Initialize RSA public key
      rsaInitPublicKey(&publicKey);

      //Decode pem key file into RSA public key
      error = pemImportRsaPublicKey(cryptoSettings->signKey, cryptoSettings->signKeyLen, &publicKey);
      
      //Check status code
      if (!error)
      {
         //Set check data (signature) size
         context->checkDataSize = publicKey.n.size * sizeof(publicKey.n.size);
   
         //Free RSA public key
         rsaFreePublicKey(&publicKey);
      }
   }
   else
#endif
#if IAP_ECDSA_SUPPORT == ENABLED
   //Is signature ECDSA algorithm?
   if (settings->signAlgo == IAP_SIGN_ECDSA)
   {
      //Initialize signature algo context
      settings->signHashAlgo->init(context->checkContext);

      //Set digest length
      context->digestLen = settings->signHashAlgo->digestSize;
      
      //Initialize EC domain parameters
      ecInitDomainParameters(&ecParams);
      
      //Load EC domain parameters
      error = ecLoadDomainParameters(&ecParams, SECP256K1_CURVE);
      
      //Check status code
      if(!error)
      {
         //Set check data (signature) size
         context->checkDataSize = mpiGetByteLength(&ecParams.q) * 2;
         
         //EC domain parameters
         ecFreeDomainParameters(&ecParams);
      }
   }
   else
#endif
   {
      //Debug message
      TRACE_ERROR("Signature algorithm not supported!\r\n");
      error = ERROR_NOT_IMPLEMENTED;
   }

   //Return status code 
   return error;
#else
   //Not implemented
   return ERROR_NOT_IMPLEMENTED;
#endif
}


/**
 * @brief Verify received image firmware signature.
 * @param[in,out] context Pointer to the IAP context
 * @return Error code
 **/

error_t iapSignVerify(ImageContext *context)
{
#if IAP_SIGNATURE_SUPPORT == ENABLED 
   error_t error;
   IapCryptoSettings* cryptoSettings;


   //Check parameters validity
   if (context == NULL)
      return ERROR_INVALID_PARAMETER;

   //Point to the IAP user crypto settings
   cryptoSettings = (IapCryptoSettings*)context->imageCryptoSettings;

   //Initialize status code
   error = NO_ERROR;

   //Check authentification hash algo
   if (cryptoSettings->signAlgo == IAP_SIGN_NONE)
      return ERROR_INVALID_VALUE;

#if IAP_RSA_SUPPORT == ENABLED
   //Is user require RSA signature?
   if (cryptoSettings->signAlgo == IAP_SIGN_RSA)
   {
      //Verify RSA signature
      error = iapSignVerifyRsa(context);
   }
   else
#endif
#if IAP_ECDSA_SUPPORT == ENABLED
   //Is user require ECDSA signature?
   if (settings->signAlgo == IAP_SIGN_ECDSA)
   {
      //Verify ECDSA signature
      error = iapSignVerifyEcdsa(context);
   }
   else
#endif
   {
      //Debug message
      TRACE_ERROR("Signature algorithm not supported!\r\n");
      error = ERROR_NOT_IMPLEMENTED;
   }

   //Return status code
   return error;
#else
   //Not implemented
   return ERROR_NOT_IMPLEMENTED;
#endif
}


///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


/**
 * @brief Verify RSA signature.
 * @param[in,out] context Pointer to the IAP context
 * @return Error code
 **/

error_t iapSignVerifyRsa(ImageContext* context)
{
#if IAP_SIGNATURE_SUPPORT == ENABLED && IAP_RSA_SUPPORT == ENABLED
   error_t error;
   IapCryptoSettings* cryptoSettings;

   RsaPublicKey publicKey;

   //Check parameter validity
   if (context == NULL)
      return ERROR_INVALID_PARAMETER;

   //Point to the IAP crypto user settings
   cryptoSettings = context->imageCryptoSettings;

   //Check signature user settings
   if (cryptoSettings->signHashAlgo == NULL || cryptoSettings->signKey == NULL ||
      cryptoSettings->signKeyLen == 0)
      return ERROR_ABORTED;

   //Initialize RSA public key
   rsaInitPublicKey(&publicKey);

   //Import PEM RSA public key
   error = pemImportRsaPublicKey(cryptoSettings->signKey, cryptoSettings->signKeyLen, &publicKey);
   //Is any error?
   if (error)
   {
      //Debug message
      TRACE_ERROR("RSA public key import failed!\r\n");
      return error;
   }

   error = rsassaPkcs1v15Verify(&publicKey, cryptoSettings->signHashAlgo, 
      context->digest, context->checkData, context->checkDataLen);
   //Is any error?
   if (error)
   {
      //Debug message
      TRACE_ERROR("RSA signature verification failed!\r\n");
      return error;
   }

   //Successful process
   return NO_ERROR;
#else
   //Not implemented
   return ERROR_NOT_IMPLEMENTED;
#endif
}


/**
 * @brief Verify ECDSA signature.
 * @param[in,out] context Pointer to the IAP context
 * @return Error code
 **/

error_t iapSignVerifyEcdsa(ImageContext* context)
{
#if IAP_SIGNATURE_SUPPORT == ENABLED && IAP_ECDSA_SUPPORT
   error_t error;
   IapCryptoSettings * *settings;

   EcDomainParameters ecParams;
   EcdsaSignature ecdsaSignature;
   EcPoint publicKey;
   uint8_t *p;

   //Intialize status code
   error = NO_ERROR;
   
   //Check parameter validity
   if (context == NULL)
      return ERROR_INVALID_PARAMETER;

   //Point to the IAP crypto user settings
   cryptoSettings = context->imageCryptoSettings;
   
   //Check signature user settings
   if (settings->signHashAlgo == NULL || settings->signKey == NULL ||
      settings->signKeyLen == 0)
      return ERROR_ABORTED;

   //Initialize EC domain parameters
   ecInitDomainParameters(&ecParams);
   //Initialize ECDSA signature
   ecdsaInitSignature(&ecdsaSignature);
   //Initialize EC public keys
   ecInit(&publicKey);
   
   //Beginning of exceptions handling
   do
   {
      //Point 
      p = (uint8_t*) context->checkData;
      
      //Import R value into ECDSA signature structure from receive check data
      error = mpiImport(&ecdsaSignature.r, p, context->checkDataLen/2, MPI_FORMAT_BIG_ENDIAN);
      //Is any error?
      if(error)
         break;
      
      p+= context->checkDataLen/2;
      
      //Import S value into ECDSA signature structure from receive check data
      error = mpiImport(&ecdsaSignature.s, p, context->checkDataLen/2, MPI_FORMAT_BIG_ENDIAN);
      //Is any error?
      if(error)
         break;
      
      //Load EC domain parameters
      error = ecLoadDomainParameters(&ecParams, SECP256K1_CURVE);
      //Is any error?
      if(error)
         break;
      
      //Decode the PEM file that contains the EC public key
      error = pemImportEcPublicKey(context->settings.signKey, context->settings.signKeyLen, 
         &publicKey);
      //Is any error?
      if(error)
         break;
      
      //Verify EDCSA signature
      error = ecdsaVerifySignature(&ecParams, &publicKey, context->digest, 
         context->digestLen, &ecdsaSignature);
      //Is any error?
      if(error)
      {
         //Debug message
         TRACE_ERROR("ECDSA signature verification failed!\r\n");
         break;
      }
   }while(0);
   
   //Release previously allocated resources
   ecFreeDomainParameters(&ecParams);
   ecdsaFreeSignature(&ecdsaSignature);
   ecFree(&publicKey);
   
   //Return Status code
   return error;
#else
   //Not implemented
   return ERROR_NOT_IMPLEMENTED;
#endif
}
