/**
 * @file uart_user.c
 * @brief User UART routine
 *
 * @section License
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 *
 * Copyright (C) 2010-2022 Oryx Embedded SARL. All rights reserved.
 *
 * This file is part of CycloneBOOT Open.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * @author Oryx Embedded SARL (www.oryx-embedded.com)
 * @version 2.0.2
 **/

//Dependencies
#include "stm32h7xx.h"
#include "stm32h7xx_hal.h"
#include "os_port.h"
#include "main.h"
#include "iap/iap.h"
#include "iap_user_settings.h"
#include "uart_user.h"
#include "common.h"
#include "ymodem.h"
#include "error.h"
#include "debug.h"

//Y-Modem uart handler
UART_HandleTypeDef UartHandle;

//CycloneBOOT Iap Context 
IapContext iapContext;
//CycloneBOOT Iap Settings
IapSettings iapSettings;

//IAP mutex handler
extern OsMutex iapMutex;

//Extern Y-Modem filename buffer holder
extern uint8_t aFileName[FILE_NAME_LENGTH];


/**
 * @brief Download a file via serial port
 **/
  
void uartUserDownload(void)
{
   uint8_t number[11] = {0};
   uint32_t size = 0;
   COM_StatusTypeDef result;
  
   error_t error;
  
   //Get exclusive access
   osAcquireMutex(&iapMutex);
   
   //Suspend scheduler
   osSuspendAllTasks();

   //Debug message
   TRACE_INFO("\r\n");
   TRACE_INFO("Starting firmware update...\r\n");
  
   //Get default IAP settings
   iapGetDefaultSettings(&iapSettings);

#if defined(IAP_SCENARIO_2)
   //User IAP settings configuration
   iapSettings.imageInCrypto.authAlgo          = IAP_SETTING_AUTH_ALGO;
   iapSettings.imageInCrypto.authHashAlgo      = IAP_SETTING_AUTH_HASH_ALGO;
   iapSettings.imageInCrypto.authKey           = IAP_SETTING_AUTH_KEY;
   iapSettings.imageInCrypto.authKeyLen        = IAP_SETTING_AUTH_KEY_LEN;
   iapSettings.imageInCrypto.cipherAlgo        = IAP_SETTING_CIPHER_ALGO;
   iapSettings.imageInCrypto.cipherMode        = IAP_SETTING_CIPHER_MODE;
   iapSettings.imageInCrypto.cipherKey         = IAP_SETTING_CIPHER_KEY;
   iapSettings.imageInCrypto.cipherKeyLen      = IAP_SETTING_CIHPER_KEY_LEN;
#elif defined(IAP_SCENARIO_3)
   //User IAP settings configuration
   iapSettings.imageInCrypto.signAlgo          = IAP_SETTING_SIGN_ALGO;
   iapSettings.imageInCrypto.signHashAlgo      = IAP_SETTING_SIGN_HASH_ALGO;
   iapSettings.imageInCrypto.signKey           = IAP_SETTING_SIGN_KEY;
   iapSettings.imageInCrypto.signKeyLen        = IAP_SETTING_SIGN_KEY_LEN;
   iapSettings.imageInCrypto.cipherAlgo        = IAP_SETTING_CIPHER_ALGO;
   iapSettings.imageInCrypto.cipherMode        = IAP_SETTING_CIPHER_MODE;
   iapSettings.imageInCrypto.cipherKey         = IAP_SETTING_CIPHER_KEY;
   iapSettings.imageInCrypto.cipherKeyLen      = IAP_SETTING_CIHPER_KEY_LEN;
#else
   //User IAP settings configuration
   iapSettings.imageInCrypto.integrityAlgo     = IAP_SETTING_INTEGRITY_ALGO;
#endif
   //User IAP primary(internal flash) memory driver 
   iapSettings.primaryFlashDriver             = IAP_SETTING_PRM_MEM_DRIVER;

   //Initialize IAP Application context
   error = iapInit(&iapContext, &iapSettings);
   //Is any error?
   if(error)
   {
      //Debug message
      TRACE_ERROR("Failed to initialize IAP!\r\n");
   }
   else
   {
      Serial_PutString((uint8_t *)"Waiting for the file to be sent ... (press 'a' to abort)\n\r");
      result = Ymodem_Receive( &size );
      if (result == COM_OK)
      {
         Serial_PutString((uint8_t *)"\n\n\r Programming Completed Successfully!\n\r--------------------------------\r\n Name: ");
         Serial_PutString(aFileName);
         Int2Str(number, size);
         Serial_PutString((uint8_t *)"\n\r Size: ");
         Serial_PutString(number);
         Serial_PutString((uint8_t *)" Bytes\r\n");
         Serial_PutString((uint8_t *)"-------------------\n");
        
         error = iapFinalize(&iapContext);
         if(error)
         {
            //Debug message
            TRACE_ERROR("Failed to finalize firmware update!\r\n");
         }
         else
         {   
            error = iapReboot(&iapContext);
            if(error)
            {
               //Debug message
               TRACE_ERROR("Failed to reboot!\r\n");
            }
         }    
      }
      else if (result == COM_LIMIT)
      {
         Serial_PutString((uint8_t *)"\n\n\rThe image size is higher than the allowed space memory!\n\r");
      }
      else if (result == COM_DATA)
      {
         Serial_PutString((uint8_t *)"\n\n\rVerification failed!\n\r");
      }
      else if (result == COM_ABORT)
      {
         Serial_PutString((uint8_t *)"\r\n\nAborted by user.\n\r");
      }
      else
      {
         Serial_PutString((uint8_t *)"\n\rFailed to receive the file!\n\r");
      }
   }
   
   //Suspend scheduler
   osResumeAllTasks();
   
   //Release exclusive access
   osReleaseMutex(&iapMutex);
}


/**
 * @brief Y-Modem Receive User callback
 * @param[in] data Pointer to the received data
 * @param[in] length Length of the received data
 * @retuen Error code
 **/

error_t Ymodem_Receive_UserCallback(uint8_t *data, size_t length)
{
   error_t error;
   
   //Write received bytes in flash
   error = iapUpdate(&iapContext, data, length);
   //Is any error?
   if(error)
   {
      //Debug message
      TRACE_ERROR("Failed to update firmware!\r\n");
   }
   
   //Status code
   return error;
}


/**
 * @brief Initialize user uart
 **/

void uartUserInit(void)
{ 
   GPIO_InitTypeDef GPIO_InitStructure;

#if (DEBUG_CONFIG == DEBUG_UART_CONFIG_2)   
   //Enable GPIOD clock
   __HAL_RCC_GPIOD_CLK_ENABLE();
   //Enable USART3 clock
   __HAL_RCC_USART3_CLK_ENABLE();

   //Configure USART3_TX (PD8)
   GPIO_InitStructure.Pin = GPIO_PIN_8;
   GPIO_InitStructure.Mode = GPIO_MODE_AF_PP;
   GPIO_InitStructure.Pull = GPIO_PULLUP;
   GPIO_InitStructure.Speed = GPIO_SPEED_FREQ_MEDIUM;
   GPIO_InitStructure.Alternate = GPIO_AF7_USART3;
   HAL_GPIO_Init(GPIOD, &GPIO_InitStructure);

   //Configure USART3_RX (PD9)
   GPIO_InitStructure.Pin = GPIO_PIN_9;
   GPIO_InitStructure.Mode = GPIO_MODE_AF_PP;
   GPIO_InitStructure.Pull = GPIO_PULLUP;
   GPIO_InitStructure.Speed = GPIO_SPEED_FREQ_MEDIUM;
   GPIO_InitStructure.Alternate = GPIO_AF7_USART3;
   HAL_GPIO_Init(GPIOD, &GPIO_InitStructure);

   //Configure USART3
   UartHandle.Instance = USART3;
   UartHandle.Init.BaudRate = 115200;
   UartHandle.Init.WordLength = UART_WORDLENGTH_8B;
   UartHandle.Init.StopBits = UART_STOPBITS_1;
   UartHandle.Init.Parity = UART_PARITY_NONE;
   UartHandle.Init.HwFlowCtl = UART_HWCONTROL_NONE;
   UartHandle.Init.Mode = UART_MODE_TX_RX;
   HAL_UART_Init(&UartHandle);   
#else
   //Enable GPIOD clock
   __HAL_RCC_GPIOD_CLK_ENABLE();
   //Enable USART2 clock
   __HAL_RCC_USART2_CLK_ENABLE();

   //Configure USART2_TX (PD5)
   GPIO_InitStructure.Pin = GPIO_PIN_5;
   GPIO_InitStructure.Mode = GPIO_MODE_AF_PP;
   GPIO_InitStructure.Pull = GPIO_PULLUP;
   GPIO_InitStructure.Speed = GPIO_SPEED_FREQ_MEDIUM;
   GPIO_InitStructure.Alternate = GPIO_AF7_USART2;
   HAL_GPIO_Init(GPIOD, &GPIO_InitStructure);

   //Configure USART2_RX (PD6)
   GPIO_InitStructure.Pin = GPIO_PIN_6;
   GPIO_InitStructure.Mode = GPIO_MODE_AF_PP;
   GPIO_InitStructure.Pull = GPIO_PULLUP;
   GPIO_InitStructure.Speed = GPIO_SPEED_FREQ_MEDIUM;
   GPIO_InitStructure.Alternate = GPIO_AF7_USART2;
   HAL_GPIO_Init(GPIOD, &GPIO_InitStructure);

   //Configure USART2
   UartHandle.Instance = USART2;
   UartHandle.Init.BaudRate = 9600;
   UartHandle.Init.WordLength = UART_WORDLENGTH_8B;
   UartHandle.Init.StopBits = UART_STOPBITS_1;
   UartHandle.Init.Parity = UART_PARITY_NONE;
   UartHandle.Init.HwFlowCtl = UART_HWCONTROL_NONE;
   UartHandle.Init.Mode = UART_MODE_TX_RX;
   HAL_UART_Init(&UartHandle);
#endif
}


/**
 * @brief Uart user task routine
 **/

void uartUserTask(void *param)
{
   uint8_t key = 0;
   uint8_t state = 0;
   HAL_StatusTypeDef status;
   
   //Init uart
   uartUserInit();

   while(1)
   {
      if(state == 0)
      {
         Serial_PutString((uint8_t*) "\r\n");
         Serial_PutString((uint8_t*) "=======================\r\n");
         Serial_PutString((uint8_t*) "= Y-MODEM update menu =\r\n");
         Serial_PutString((uint8_t*) "=======================\r\n");
         Serial_PutString((uint8_t*) "Choices:\r\n");
         Serial_PutString((uint8_t*) "  1 -> Download update image\r\n");
         Serial_PutString((uint8_t*) "  2 -> Get current application version\r\n");
         Serial_PutString((uint8_t*) "\r\n");
         
         //Change state
         state = 1;
      }
      else if(state == 1)
      {
         /* Clean the input path */
         __HAL_UART_FLUSH_DRREGISTER(&UartHandle);

         /* Receive key */
         status = HAL_UART_Receive(&UartHandle, &key, 1, RX_TIMEOUT);
         
         if(status == HAL_OK)
         {
            switch(key)
            {
               case '1':
                  Serial_PutString((uint8_t*) "Downloading update image...\r\n");
                  //Start downloading file
                  uartUserDownload();
                  break;
               
               case '2':
#if defined(APP_DEMO_RED)
                  Serial_PutString((uint8_t*) "RED (v1.0.0)\r\n");
#elif defined(APP_DEMO_GREEN)
                  Serial_PutString((uint8_t*) "GREEN (v2.0.0)\r\n");
#else
                  Serial_PutString((uint8_t*) "YELLOW (v3.0.0)\r\n");
#endif
                  break;
               
               default:
                  Serial_PutString((uint8_t*) "Unknown number!\r\n");
                  Serial_PutString((uint8_t*) "Number should be 1 or 2!\r\n");
                  break;
            }
            Serial_PutString((uint8_t*) "\r\n");
            
            //Change state
            state = 0;
         }
      }
      else
      {
         //For sanity
      }
      
      //Os delay
      osDelayTask(200);
   }
}
