/**
 * @file main.c
 * @brief Main routine
 *
 * @section License
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 *
 * Copyright (C) 2010-2022 Oryx Embedded SARL. All rights reserved.
 *
 * This file is part of CycloneBOOT Open.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * @author Oryx Embedded SARL (www.oryx-embedded.com)
 * @version 2.0.2
 **/

//Dependencies
#include <stdlib.h>
#include "stm32l4xx.h"
#include "stm32l4xx_hal.h"
#include "stm32l4xx_nucleo_144.h"
#include "usbd_core.h"
#include "core/net.h"
#include "drivers/usb_rndis/rndis_driver.h"
#include "drivers/usb_rndis/usbd_desc.h"
#include "drivers/usb_rndis/usbd_rndis.h"
#include "dhcp/dhcp_server.h"
#include "ipv6/ndp_router_adv.h"
#include "http/http_server.h"
#include "http/mime.h"
#include "path.h"
#include "date_time.h"
#include "resource_manager.h"
#if (IAP_FALLBACK_SUPPORT == ENABLED)
#include "iap/iap_fallback.h"
#endif
#include "http_server_callbacks.h"
#include "iap_user_settings.h"
#include "version.h"
#include "debug.h"

//USB/RNDIS interface configuration
#define APP_IF_NAME "usb0"
#define APP_HOST_NAME "http-server-demo"
#define APP_MAC_ADDR "00-00-00-00-04-96"

#define APP_USE_DHCP_SERVER ENABLED
#define APP_IPV4_HOST_ADDR "192.168.9.1"
#define APP_IPV4_SUBNET_MASK "255.255.255.0"
#define APP_IPV4_DEFAULT_GATEWAY "0.0.0.0"
#define APP_IPV4_PRIMARY_DNS "0.0.0.0"
#define APP_IPV4_SECONDARY_DNS "0.0.0.0"
#define APP_IPV4_ADDR_RANGE_MIN "192.168.9.10"
#define APP_IPV4_ADDR_RANGE_MAX "192.168.9.99"

#define APP_USE_ROUTER_ADV ENABLED
#define APP_IPV6_LINK_LOCAL_ADDR "fe80::496"
#define APP_IPV6_PREFIX "fd00:1:2:3::"
#define APP_IPV6_PREFIX_LENGTH 64
#define APP_IPV6_GLOBAL_ADDR "fd00:1:2:3::496"

//Application configuration
#define APP_HTTP_MAX_CONNECTIONS 4

//#define TEST_MODE
#if defined(TEST_MODE)
//#define TEST_FLASH_R_W //uncomment if flash read write test or leave it commented to test flash swap banks
#define TEST_LED 0 // 0 green led, 1 blue led
#endif

//Global variables
USBD_HandleTypeDef USBD_Device;

DhcpServerSettings dhcpServerSettings;
DhcpServerContext dhcpServerContext;
NdpRouterAdvSettings ndpRouterAdvSettings;
NdpRouterAdvPrefixInfo ndpRouterAdvPrefixInfo[1];
NdpRouterAdvContext ndpRouterAdvContext;
HttpServerSettings httpServerSettings;
HttpServerContext httpServerContext;
HttpConnection httpConnections[APP_HTTP_MAX_CONNECTIONS];

const uint8_t* pemIapSignPublicKey;
size_t pemIapSignPublicKeyLen;

//IAP mutex handler
OsMutex iapMutex;

#if (IAP_FALLBACK_SUPPORT == ENABLED)
//User fallback routines
void fallbackUserTriggerInit(void);
error_t fallbackUserRoutine(void);
#endif


/**
 * @brief System clock configuration
 **/

void SystemClock_Config(void)
{
   RCC_OscInitTypeDef RCC_OscInitStruct = {0};
   RCC_ClkInitTypeDef RCC_ClkInitStruct = {0};
   RCC_PeriphCLKInitTypeDef RCC_PeriphClkInitStruct = {0};

   //Enable the LSE Oscilator
   RCC_OscInitStruct.OscillatorType = RCC_OSCILLATORTYPE_HSI48;
   RCC_OscInitStruct.HSI48State = RCC_HSI48_ON;
   RCC_OscInitStruct.PLL.PLLState = RCC_PLL_OFF;
   HAL_RCC_OscConfig(&RCC_OscInitStruct);

   //Enable MSI Oscillator and activate PLL with MSI as source
   RCC_OscInitStruct.OscillatorType = RCC_OSCILLATORTYPE_MSI;
   RCC_OscInitStruct.MSIState = RCC_MSI_ON;
   RCC_OscInitStruct.MSIClockRange = RCC_MSIRANGE_7;
   RCC_OscInitStruct.MSICalibrationValue = RCC_MSICALIBRATION_DEFAULT;
   RCC_OscInitStruct.PLL.PLLState = RCC_PLL_ON;
   RCC_OscInitStruct.PLL.PLLSource = RCC_PLLSOURCE_MSI;
   RCC_OscInitStruct.PLL.PLLM = 1;
   RCC_OscInitStruct.PLL.PLLN = 20;
   RCC_OscInitStruct.PLL.PLLR = 2;
   RCC_OscInitStruct.PLL.PLLP = 7;
   RCC_OscInitStruct.PLL.PLLQ = 4;
   HAL_RCC_OscConfig(&RCC_OscInitStruct);

   //Enable MSI Auto-calibration through LSE
   HAL_RCCEx_EnableMSIPLLMode();

   //Select PLL as system clock source and configure the HCLK, PCLK1 and PCLK2
   //clocks dividers
   RCC_ClkInitStruct.ClockType = RCC_CLOCKTYPE_SYSCLK | RCC_CLOCKTYPE_HCLK |
      RCC_CLOCKTYPE_PCLK1 | RCC_CLOCKTYPE_PCLK2;
   RCC_ClkInitStruct.SYSCLKSource = RCC_SYSCLKSOURCE_PLLCLK;
   RCC_ClkInitStruct.AHBCLKDivider = RCC_SYSCLK_DIV1;
   RCC_ClkInitStruct.APB1CLKDivider = RCC_HCLK_DIV1;
   RCC_ClkInitStruct.APB2CLKDivider = RCC_HCLK_DIV1;
   HAL_RCC_ClockConfig(&RCC_ClkInitStruct, FLASH_LATENCY_4);

   //Select clock source for RNG peripheral
   RCC_PeriphClkInitStruct.PeriphClockSelection = RCC_PERIPHCLK_RNG;
   RCC_PeriphClkInitStruct.RngClockSelection = RCC_RNGCLKSOURCE_HSI48;
   HAL_RCCEx_PeriphCLKConfig(&RCC_PeriphClkInitStruct);

   //Select clock source for USB peripheral
   RCC_PeriphClkInitStruct.PeriphClockSelection = RCC_PERIPHCLK_USB;
   RCC_PeriphClkInitStruct.UsbClockSelection = RCC_USBCLKSOURCE_HSI48;
   HAL_RCCEx_PeriphCLKConfig(&RCC_PeriphClkInitStruct);
}


/**
 * @brief LED task
 * @param[in] param Unused parameter
 **/

void ledTask(void *param)
{
   //Endless loop
   while(1)
   {
      BSP_LED_On(LED1);
      osDelayTask(100);
      BSP_LED_Off(LED1);
      osDelayTask(900);
   }
}


/**
 * @brief Load user RSA public key from "res.c".
 * "res.c" is a C array file system representing the demo folder "resources/keys".
 **/

error_t iapUserLoadKeys(void)
{
   error_t error;
   
   //Get server's private key
   error = resGetData(IAP_PUBLIC_SIGN_KEY, (const uint8_t **) &pemIapSignPublicKey,
      &pemIapSignPublicKeyLen);
   //Any error to report?
   if(error)
      return error;
   
   //Successful process
   return NO_ERROR;
}


#if (IAP_FALLBACK_SUPPORT == ENABLED)
/**
 * @brief Initialize the User fallback trigger
 * @return None
 **/

void fallbackUserTriggerInit(void)
{
   //Setup User button (B1 : PC13)

   // Enable the clock for GPIOC
   RCC->AHB2ENR |= (1 << 2);
   // Set PC13 as input
   GPIOC->MODER &= ~(1<<(13*2));
   GPIOC->MODER &= ~(1<<(13*2+1));

   // Set PC13 as Pull-Down
   GPIOC->PUPDR |= (2<<(13*2));
}


/**
 * @brief Fallback User routine. It executes the CycloneBOOT IAP
 * fallback according the defined User trigger action.
 * @return Status code
 **/

error_t fallbackUserRoutine(void)
{
   error_t error;

   //Initialize status code
   error = NO_ERROR;

   //Initialize the fallback trigger (here User Button -> B1 : PC13)
   fallbackUserTriggerInit();

   //Is fallback trigger hit?
   if ((GPIOC->IDR >> 13) & 0x1)
   {
      //wait button release
      while ((GPIOC->IDR >> 13) & 0x1);

      //Start CycloneBOOT IAP Fallback procedure
      error = iapFallbackStart();
   }

   //Return status code
   return error;
}
#endif


/**
 * @brief Main entry point
 * @return Unused value
 **/

int_t main(void)
{
   error_t error;
   OsTaskId taskId;
   NetInterface *interface;
   MacAddr macAddr;
#if (IPV4_SUPPORT == ENABLED)
   Ipv4Addr ipv4Addr;
#endif
#if (IPV6_SUPPORT == ENABLED)
   Ipv6Addr ipv6Addr;
#endif
   
#if (IAP_FALLBACK_SUPPORT == ENABLED)
   //Call user fallback routine
   fallbackUserRoutine();
#endif

   //HAL library initialization
   HAL_Init();
   //Configure the system clock
   SystemClock_Config();

   //Initialize kernel
   osInitKernel();
   //Configure debug UART
   debugInit(115200);

   //Start-up message
   TRACE_INFO("\r\n");
   TRACE_INFO("************************************************************\r\n");
   TRACE_INFO("*** CycloneBOOT USB/RNDIS HTTP Server Iap Dual Bank Demo ***\r\n");
   TRACE_INFO("************************************************************\r\n");
#if defined(IAP_SCENARIO_1)
   TRACE_INFO("IAP mode: Simple APP + Integrity\r\n");
#elif defined (IAP_SCENARIO_2)
   TRACE_INFO("IAP mode: Cipher APP + Authentification\r\n");
#elif defined(IAP_SCENARIO_3)
   TRACE_INFO("IAP mode: Cipher APP + Signature\r\n");
#else
   #endif
   TRACE_INFO("Version: %s\r\n", APP_VERSION_STRING);
   TRACE_INFO("Copyright: 2010-2022 Oryx Embedded SARL\r\n");
   TRACE_INFO("Compiled: %s %s\r\n", __DATE__, __TIME__);
   TRACE_INFO("Target: STM32L496\r\n");
   TRACE_INFO("\r\n");
  
   TRACE_INFO("\r\n");
   TRACE_INFO("''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''\r\n");
   TRACE_INFO("Goto \"http://" APP_IPV4_HOST_ADDR "\"\r\n");
   TRACE_INFO("Or use board hostname and goto \"http://" APP_HOST_NAME "\"\r\n");
   TRACE_INFO("''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''\r\n");
   TRACE_INFO("\r\n");

   //LED configuration
   BSP_LED_Init(LED1);
   BSP_LED_Init(LED2);
   BSP_LED_Init(LED3);

   //Clear LEDs
   BSP_LED_Off(LED1);
   BSP_LED_Off(LED2);
   BSP_LED_Off(LED3);

   //Initialize user button
   BSP_PB_Init(BUTTON_KEY, BUTTON_MODE_GPIO);

#if defined(TEST_MODE)
      {
      #include "drivers/flash/stm32l4xx_flash_driver.h"
      
      uint_t testIndex;
      uint8_t dataR[128];
      uint8_t dataW[128];
      size_t writeSize;
      uint32_t eraseStartAddr;
      uint32_t addrTest;
      size_t eraseLength;
      error_t error;
      FlashDriver *flashDrv = (FlashDriver*) &stm32l4xxFlashDriver;
      const FlashInfo *flashInfo;
      uint_t i;
      
      error = NO_ERROR;
      testIndex = 1;
      addrTest = STM32L4xx_FLASH_BANK2_ADDR;
      
      //Init flash driver
      flashDrv->init();
      
      error = flashDrv->getInfo(&flashInfo);
      if(error)
      {
         TRACE_ERROR("Failed to read flash!\r\n");
         return error;
      }
#if defined(TEST_FLASH_R_W)
      
      /////////////////////////////////////////////////////////////////////////
      /////////////////////////////////////////////////////////////////////////
      /////////////////////////////////////////////////////////////////////////

      
      TRACE_INFO("\r\n");
      TRACE_INFO("TEST%d: Read 1 byte\r\n", testIndex);
      
      do {
         error = NO_ERROR;       
         memset(dataR, 0, sizeof(dataR));
         
         //Read first bank 2 sector 128bytes
         error = flashDrv->read(STM32L4xx_FLASH_BANK2_ADDR, dataR, sizeof(uint8_t));
         if(error)
         {
            TRACE_ERROR("Failed to read flash!\r\n");
            break;
         }
      
         TRACE_INFO_ARRAY("READ: ", dataR, sizeof(uint8_t));
      }while(0);
      
      if(error)
         TRACE_INFO("TEST%d: Failed!\r\n",testIndex);
      else
         TRACE_INFO("TEST%d: OK\r\n",testIndex);
      
      testIndex++;
      
      /////////////////////////////////////////////////////////////////////////
      /////////////////////////////////////////////////////////////////////////
      /////////////////////////////////////////////////////////////////////////
      
      TRACE_INFO("\r\n");
      TRACE_INFO("TEST%d: Read 1 word\r\n",testIndex);
      
      do {
         error = NO_ERROR;       
         memset(dataR, 0, sizeof(dataR));
         
         //Read first bank 2 sector 128bytes
         error = flashDrv->read(STM32L4xx_FLASH_BANK2_ADDR, dataR, sizeof(uint32_t));
         if(error)
         {
            TRACE_ERROR("Failed to read flash!\r\n");
            break;
         }
         
         TRACE_INFO_ARRAY("READ: ", dataR, sizeof(uint32_t));
      }while(0);
      
      if(error)
         TRACE_INFO("TEST%d: Failed!\r\n",testIndex);
      else
         TRACE_INFO("TEST%d: OK\r\n",testIndex);
      
      testIndex++;
      
      /////////////////////////////////////////////////////////////////////////
      /////////////////////////////////////////////////////////////////////////
      /////////////////////////////////////////////////////////////////////////
      
      TRACE_INFO("\r\n");
      TRACE_INFO("TEST%d: Read 128 bytes\r\n",testIndex);
      
      do {
         error = NO_ERROR;       
         memset(dataR, 0, sizeof(dataR));
         
         //Read first bank 2 sector 128bytes
         error = flashDrv->read(STM32L4xx_FLASH_BANK2_ADDR, dataR, sizeof(dataR));
         if(error)
         {
            TRACE_ERROR("Failed to read flash!\r\n");
            break;
         }
         
         TRACE_INFO_ARRAY("READ: ", dataR, sizeof(dataR));
      }while(0);
      
      if(error)
         TRACE_INFO("TEST%d: Failed!\r\n", testIndex);
      else
         TRACE_INFO("TEST%d: OK\r\n",testIndex);
      
      testIndex++;
      
      /////////////////////////////////////////////////////////////////////////
      /////////////////////////////////////////////////////////////////////////
      /////////////////////////////////////////////////////////////////////////
      
      TRACE_INFO("\r\n");
      TRACE_INFO("TEST%d: Write 1 byte\r\n", testIndex);
      
      do {
         error = NO_ERROR;       
         memset(dataW, 0, sizeof(dataW));
         memset(dataW, 0xAA, sizeof(uint8_t));
         
         writeSize = (sizeof(uint8_t)*flashInfo->writeSize)/flashInfo->writeSize + ((sizeof(uint8_t) % flashInfo->writeSize) > 0);
         
         TRACE_INFO_ARRAY("WRITE: ", dataW, writeSize);
         
         error = flashDrv->write(STM32L4xx_FLASH_BANK2_ADDR, dataW, sizeof(uint8_t));
         if(error)
         {
            TRACE_ERROR("Failed to write flash!\r\n");
            break;
         }
         
         //Read first bank 2 sector 128bytes
         error = flashDrv->read(STM32L4xx_FLASH_BANK2_ADDR, dataR, sizeof(dataR));
         if(error)
         {
            TRACE_ERROR("Failed to read flash!\r\n");
            break;
         }
   
         TRACE_INFO_ARRAY("READ: ", dataR, writeSize);
         
         if(memcmp(dataW, dataR, writeSize) != 0)
         {
            error = ERROR_FAILURE;
            break;
         }
      }while(0);
      
      if(error)
         TRACE_INFO("TEST%d: Failed!\r\n",testIndex);
      else
         TRACE_INFO("TEST%d: OK\r\n",testIndex);
      
      testIndex++;
      
      /////////////////////////////////////////////////////////////////////////
      /////////////////////////////////////////////////////////////////////////
      /////////////////////////////////////////////////////////////////////////
      
      TRACE_INFO("\r\n");
      TRACE_INFO("TEST%d: Write 1 word\r\n",testIndex);
      
      do {
         error = NO_ERROR;       
         memset(dataW, 0, sizeof(dataW));
         memset(dataW, 0xBB, sizeof(uint32_t));
         
         writeSize = (sizeof(uint32_t)*flashInfo->writeSize)/flashInfo->writeSize + ((sizeof(uint32_t) % flashInfo->writeSize) > 0);
         
         TRACE_INFO_ARRAY("WRITE: ", dataW, writeSize);
         
         error = flashDrv->write(STM32L4xx_FLASH_BANK2_ADDR, dataW, sizeof(uint32_t));
         if(error)
         {
            TRACE_ERROR("Failed to write flash!\r\n");
            break;
         }
         
         //Read first bank 2 sector 128bytes
         error = flashDrv->read(STM32L4xx_FLASH_BANK2_ADDR, dataR, sizeof(dataR));
         if(error)
         {
            TRACE_ERROR("Failed to read flash!\r\n");
            break;
         }
   
         TRACE_INFO_ARRAY("READ: ", dataR, writeSize);
         
         if(memcmp(dataW, dataR, writeSize) != 0)
         {
            error = ERROR_FAILURE;
            break;
         }
      }while(0);
      
      if(error)
         TRACE_INFO("TEST%d: Failed!\r\n",testIndex);
      else
         TRACE_INFO("TEST%d: OK\r\n",testIndex);
      
      testIndex++;
      
      /////////////////////////////////////////////////////////////////////////
      /////////////////////////////////////////////////////////////////////////
      /////////////////////////////////////////////////////////////////////////
      
      TRACE_INFO("\r\n");
      TRACE_INFO("TEST%d: Write 8 words\r\n",testIndex);
      
      do {
         error = NO_ERROR;       
         memset(dataW, 0, sizeof(dataW));
         memset(dataW, 0xCC, flashInfo->writeSize);
         
         writeSize = (flashInfo->writeSize*flashInfo->writeSize)/flashInfo->writeSize + ((flashInfo->writeSize % flashInfo->writeSize) > 0);
         
         TRACE_INFO_ARRAY("WRITE: ", dataW, writeSize);
         
         error = flashDrv->write(STM32L4xx_FLASH_BANK2_ADDR, dataW, flashInfo->writeSize);
         if(error)
         {
            TRACE_ERROR("Failed to write flash!\r\n");
            break;
         }
         
         //Read first bank 2 sector 128bytes
         error = flashDrv->read(STM32L4xx_FLASH_BANK2_ADDR, dataR, sizeof(dataR));
         if(error)
         {
            TRACE_ERROR("Failed to read flash!\r\n");
            break;
         }
   
         TRACE_INFO_ARRAY("READ: ", dataR, writeSize);
         
         if(memcmp(dataW, dataR, writeSize) != 0)
         {
            error = ERROR_FAILURE;
            break;
         }
      }while(0);
      
      if(error)
         TRACE_INFO("TEST%d: Failed!\r\n",testIndex);
      else
         TRACE_INFO("TEST%d: OK\r\n",testIndex);
      
      testIndex++;
      
      /////////////////////////////////////////////////////////////////////////
      /////////////////////////////////////////////////////////////////////////
      /////////////////////////////////////////////////////////////////////////
      
      TRACE_INFO("\r\n");
      TRACE_INFO("TEST%d: Write 128 words\r\n",testIndex);
      
      do {
         error = NO_ERROR;       
         memset(dataW, 0, sizeof(dataW));
         memset(dataW, 0xDD, sizeof(dataW));
         
         writeSize = (sizeof(dataW)*flashInfo->writeSize)/flashInfo->writeSize + ((sizeof(dataW) % flashInfo->writeSize) > 0);
         
         TRACE_INFO_ARRAY("WRITE: ", dataW, writeSize);
         
         error = flashDrv->write(STM32L4xx_FLASH_BANK2_ADDR, dataW, sizeof(dataW));
         if(error)
         {
            TRACE_ERROR("Failed to write flash!\r\n");
            break;
         }
         
         //Read first bank 2 sector 128bytes
         error = flashDrv->read(STM32L4xx_FLASH_BANK2_ADDR, dataR, sizeof(dataR));
         if(error)
         {
            TRACE_ERROR("Failed to read flash!\r\n");
            break;
         }
   
         TRACE_INFO_ARRAY("READ: ", dataR, writeSize);
         
         if(memcmp(dataW, dataR, writeSize) != 0)
         {
            error = ERROR_FAILURE;
            break;
         }
      }while(0);
      
      if(error)
         TRACE_INFO("TEST%d: Failed!\r\n",testIndex);
      else
         TRACE_INFO("TEST%d: OK\r\n",testIndex);
      
      testIndex++;
      
      /////////////////////////////////////////////////////////////////////////
      /////////////////////////////////////////////////////////////////////////
      /////////////////////////////////////////////////////////////////////////
      
      TRACE_INFO("\r\n");
      TRACE_INFO("TEST%d: Erase flash bank 1 last sector\r\n",testIndex);
      
      do {
         error = NO_ERROR;       
         memset(dataW, 0, sizeof(dataW));
         memset(dataW, 0xEE, sizeof(dataW));
         
         eraseStartAddr = STM32L4xx_FLASH_BANK1_ADDR+STM32L4xx_FLASH_SECTOR_SIZE*(STM32L4xx_FLASH_SECTOR_NUMBER/2-1);
         eraseLength = STM32L4xx_FLASH_SECTOR_SIZE;
         
         writeSize = sizeof(dataW);
         
         TRACE_INFO_ARRAY("WRITE: ", dataW, writeSize);
         
         error = flashDrv->write(eraseStartAddr, dataW, writeSize);
         if(error)
         {
            TRACE_ERROR("Failed to write flash!\r\n");
            break;
         }
         
         error = flashDrv->erase(eraseStartAddr, eraseLength);
         if(error)
         {
            TRACE_ERROR("Failed to erase flash sector(s)!\r\n");
            break;
         }
         
         
         //Read first bank 2 sector 128bytes
         error = flashDrv->read(eraseStartAddr, dataR, writeSize);
         if(error)
         {
            TRACE_ERROR("Failed to read flash!\r\n");
            break;
         }
   
         TRACE_INFO_ARRAY("READ: ", dataR, writeSize);
         
         memset(dataW, 0xFF, writeSize);
         if(memcmp(dataW, dataR, writeSize) != 0)
         {
            error = ERROR_FAILURE;
            break;
         }
      }while(0);
      
      if(error)
         TRACE_INFO("TEST%d: Failed!\r\n",testIndex);
      else
         TRACE_INFO("TEST%d: OK\r\n",testIndex);
      
      testIndex++;
      
      /////////////////////////////////////////////////////////////////////////
      /////////////////////////////////////////////////////////////////////////
      /////////////////////////////////////////////////////////////////////////
      
      TRACE_INFO("\r\n");
      TRACE_INFO("TEST%d: Erase flash bank 2 last sector\r\n",testIndex);
      
      do {
         error = NO_ERROR;       
         memset(dataW, 0, sizeof(dataW));
         memset(dataW, 0xEE, sizeof(dataW));
         
         eraseStartAddr = STM32L4xx_FLASH_BANK2_ADDR+STM32L4xx_FLASH_SECTOR_SIZE*(STM32L4xx_FLASH_SECTOR_NUMBER/2-1);
         eraseLength = STM32L4xx_FLASH_SECTOR_SIZE;
         
         writeSize = sizeof(dataW);
         
         TRACE_INFO_ARRAY("WRITE: ", dataW, writeSize);
         
         error = flashDrv->write(eraseStartAddr, dataW, writeSize);
         if(error)
         {
            TRACE_ERROR("Failed to write flash!\r\n");
            break;
         }
         
         error = flashDrv->erase(eraseStartAddr, eraseLength);
         if(error)
         {
            TRACE_ERROR("Failed to erase flash sector(s)!\r\n");
            break;
         }
         
         
         //Read first bank 2 sector 128bytes
         error = flashDrv->read(eraseStartAddr, dataR, writeSize);
         if(error)
         {
            TRACE_ERROR("Failed to read flash!\r\n");
            break;
         }
   
         TRACE_INFO_ARRAY("READ: ", dataR, writeSize);
         
         memset(dataW, 0xFF, writeSize);
         if(memcmp(dataW, dataR, writeSize) != 0)
         {
            error = ERROR_FAILURE;
            break;
         }
      }while(0);
      
      if(error)
         TRACE_INFO("TEST%d: Failed!\r\n",testIndex);
      else
         TRACE_INFO("TEST%d: OK\r\n",testIndex);
      
      testIndex++;
         
      /////////////////////////////////////////////////////////////////////////
      /////////////////////////////////////////////////////////////////////////
      /////////////////////////////////////////////////////////////////////////
      
      TRACE_INFO("\r\n");
      TRACE_INFO("TEST%d: Erase 3 sector in flash bank 2\r\n",testIndex);
      
      do {
         error = NO_ERROR;       
         memset(dataW, 0, sizeof(dataW));
         memset(dataW, 0xEE, sizeof(dataW));
         
         eraseStartAddr = STM32L4xx_FLASH_BANK2_ADDR;
         eraseLength = STM32L4xx_FLASH_SECTOR_SIZE*3;
         
         writeSize = sizeof(dataW);
         
         TRACE_INFO_ARRAY("WRITE: ", dataW, writeSize);
         
         for(i = 0; i < 3; i++)
         {
            //write in sectors
            error = flashDrv->write(eraseStartAddr+i*STM32L4xx_FLASH_SECTOR_SIZE, dataW, writeSize);
            if(error)
            {
               TRACE_ERROR("Failed to write flash!\r\n");
               break;
            }
         }
         if(error)
            break;
         
         error = flashDrv->erase(eraseStartAddr, eraseLength);
         if(error)
         {
            TRACE_ERROR("Failed to erase flash sector(s)!\r\n");
            break;
         }
         
         //SCB_InvalidateICache();
         //SCB_InvalidateDCache();
         
         memset(dataW, 0xFF, writeSize);
         
         for(i = 0; i < 3; i++)
         {
            memset(dataR, 0x00, writeSize);
            //Read bank 1
            error = flashDrv->read(eraseStartAddr+i*STM32L4xx_FLASH_SECTOR_SIZE, dataR, writeSize);
            if(error)
            {
               TRACE_ERROR("Failed to read flash!\r\n");
               break;
            }
      
            TRACE_INFO_ARRAY("READ: ", dataR, writeSize);
            
            if(memcmp(dataW, dataR, writeSize) != 0)
            {
               error = ERROR_FAILURE;
               break;
            }
         }
         
      }while(0);
      
      if(error)
         TRACE_INFO("TEST%d: Failed!\r\n",testIndex);
      else
         TRACE_INFO("TEST%d: OK\r\n",testIndex);
      
      testIndex++;
      
      
      #else  
      
      /////////////////////////////////////////////////////////////////////////
      /////////////////////////////////////////////////////////////////////////
      /////////////////////////////////////////////////////////////////////////
      
      
      TRACE_INFO("\r\n");
      TRACE_INFO("TEST%d: Swap flash banks\r\n",testIndex);
      
      do {
         
#if (defined(TEST_LED) && (TEST_LED == 0))
         BSP_LED_On(LED1);
         BSP_LED_Off(LED2);
#else
         BSP_LED_Off(LED1);
         BSP_LED_On(LED2);
#endif         
         TRACE_INFO("Push user button to swap banks\r\n");
         
         while (BSP_PB_GetState(BUTTON_USER) != GPIO_PIN_SET);
         while (BSP_PB_GetState(BUTTON_USER) == GPIO_PIN_SET);
            
         
         error = flashDrv->swapBanks();
         if(error)
         {
            TRACE_ERROR("Failed to swap banks!\r\n");
            break;
         }
         
         NVIC_SystemReset();
      }while(0);
      
      if(error)
         TRACE_INFO("TEST%d: Failed!\r\n",testIndex);
      else
         TRACE_INFO("TEST%d: OK\r\n",testIndex);
      
      testIndex++;
      #endif
      
      
      
      
      return 0;
      
   }
#endif

   //Initialize USB device library
   USBD_Init(&USBD_Device, &usbdRndisDescriptors, 0);
   //Register RNDIS class driver
   USBD_RegisterClass(&USBD_Device, USBD_RNDIS_CLASS);
   //Start USB device core
   USBD_Start(&USBD_Device);

   //TCP/IP stack initialization
   error = netInit();
   //Any error to report?
   if(error)
   {
      //Debug message
      TRACE_ERROR("Failed to initialize TCP/IP stack!\r\n");
   }

   //Configure USB/RNDIS interface
   interface = &netInterface[0];

   //Set interface name
   netSetInterfaceName(interface, APP_IF_NAME);
   //Set host name
   netSetHostname(interface, APP_HOST_NAME);
   //Set host MAC address
   macStringToAddr(APP_MAC_ADDR, &macAddr);
   netSetMacAddr(interface, &macAddr);
   //Select the relevant network adapter
   netSetDriver(interface, &rndisDriver);

   //Initialize network interface
   error = netConfigInterface(interface);
   //Any error to report?
   if(error)
   {
      //Debug message
      TRACE_ERROR("Failed to configure interface %s!\r\n", interface->name);
   }

#if (IPV4_SUPPORT == ENABLED)
   //Set IPv4 host address
   ipv4StringToAddr(APP_IPV4_HOST_ADDR, &ipv4Addr);
   ipv4SetHostAddr(interface, ipv4Addr);

   //Set subnet mask
   ipv4StringToAddr(APP_IPV4_SUBNET_MASK, &ipv4Addr);
   ipv4SetSubnetMask(interface, ipv4Addr);

   //Set default gateway
   ipv4StringToAddr(APP_IPV4_DEFAULT_GATEWAY, &ipv4Addr);
   ipv4SetDefaultGateway(interface, ipv4Addr);

   //Set primary and secondary DNS servers
   ipv4StringToAddr(APP_IPV4_PRIMARY_DNS, &ipv4Addr);
   ipv4SetDnsServer(interface, 0, ipv4Addr);
   ipv4StringToAddr(APP_IPV4_SECONDARY_DNS, &ipv4Addr);
   ipv4SetDnsServer(interface, 1, ipv4Addr);

#if (APP_USE_DHCP_SERVER == ENABLED)
   //Get default settings
   dhcpServerGetDefaultSettings(&dhcpServerSettings);
   //Set the network interface to be configured by DHCP
   dhcpServerSettings.interface = interface;
   //Lease time, in seconds, assigned to the DHCP clients
   dhcpServerSettings.leaseTime = 3600;

   //Lowest and highest IP addresses in the pool that are available
   //for dynamic address assignment
   ipv4StringToAddr(APP_IPV4_ADDR_RANGE_MIN, &dhcpServerSettings.ipAddrRangeMin);
   ipv4StringToAddr(APP_IPV4_ADDR_RANGE_MAX, &dhcpServerSettings.ipAddrRangeMax);

   //Subnet mask
   ipv4StringToAddr(APP_IPV4_SUBNET_MASK, &dhcpServerSettings.subnetMask);
   //Default gateway
   ipv4StringToAddr(APP_IPV4_DEFAULT_GATEWAY, &dhcpServerSettings.defaultGateway);
   //DNS servers
   ipv4StringToAddr(APP_IPV4_PRIMARY_DNS, &dhcpServerSettings.dnsServer[0]);
   ipv4StringToAddr(APP_IPV4_SECONDARY_DNS, &dhcpServerSettings.dnsServer[1]);

   //DHCP server initialization
   error = dhcpServerInit(&dhcpServerContext, &dhcpServerSettings);
   //Failed to initialize DHCP client?
   if(error)
   {
      //Debug message
      TRACE_ERROR("Failed to initialize DHCP server!\r\n");
   }

   //Start DHCP server
   error = dhcpServerStart(&dhcpServerContext);
   //Failed to start DHCP client?
   if(error)
   {
      //Debug message
      TRACE_ERROR("Failed to start DHCP server!\r\n");
   }
#endif
#endif

#if (IPV6_SUPPORT == ENABLED)
   //Set link-local address
   ipv6StringToAddr(APP_IPV6_LINK_LOCAL_ADDR, &ipv6Addr);
   ipv6SetLinkLocalAddr(interface, &ipv6Addr);

   //Set IPv6 prefix
   ipv6StringToAddr(APP_IPV6_PREFIX, &ipv6Addr);
   ipv6SetPrefix(interface, 0, &ipv6Addr, APP_IPV6_PREFIX_LENGTH);

   //Set global address
   ipv6StringToAddr(APP_IPV6_GLOBAL_ADDR, &ipv6Addr);
   ipv6SetGlobalAddr(interface, 0, &ipv6Addr);

#if (APP_USE_ROUTER_ADV == ENABLED)
   //List of IPv6 prefixes to be advertised
   ipv6StringToAddr(APP_IPV6_PREFIX, &ndpRouterAdvPrefixInfo[0].prefix);
   ndpRouterAdvPrefixInfo[0].length = APP_IPV6_PREFIX_LENGTH;
   ndpRouterAdvPrefixInfo[0].onLinkFlag = TRUE;
   ndpRouterAdvPrefixInfo[0].autonomousFlag = TRUE;
   ndpRouterAdvPrefixInfo[0].validLifetime = 3600;
   ndpRouterAdvPrefixInfo[0].preferredLifetime = 1800;

   //Get default settings
   ndpRouterAdvGetDefaultSettings(&ndpRouterAdvSettings);
   //Set the underlying network interface
   ndpRouterAdvSettings.interface = interface;
   //Maximum time interval between unsolicited Router Advertisements
   ndpRouterAdvSettings.maxRtrAdvInterval = 60000;
   //Minimum time interval between unsolicited Router Advertisements
   ndpRouterAdvSettings.minRtrAdvInterval = 20000;
   //Router lifetime
   ndpRouterAdvSettings.defaultLifetime = 0;
   //List of IPv6 prefixes
   ndpRouterAdvSettings.prefixList = ndpRouterAdvPrefixInfo;
   ndpRouterAdvSettings.prefixListLength = arraysize(ndpRouterAdvPrefixInfo);

   //RA service initialization
   error = ndpRouterAdvInit(&ndpRouterAdvContext, &ndpRouterAdvSettings);
   //Failed to initialize DHCPv6 client?
   if(error)
   {
      //Debug message
      TRACE_ERROR("Failed to initialize RA service!\r\n");
   }

   //Start RA service
   error = ndpRouterAdvStart(&ndpRouterAdvContext);
   //Failed to start RA service?
   if(error)
   {
      //Debug message
      TRACE_ERROR("Failed to start RA service!\r\n");
   }
#endif
#endif

   //Get default settings
   httpServerGetDefaultSettings(&httpServerSettings);
   //Bind HTTP server to the desired interface
   httpServerSettings.interface = &netInterface[0];
   //Listen to port 80
   httpServerSettings.port = HTTP_PORT;
   //Client connections
   httpServerSettings.maxConnections = APP_HTTP_MAX_CONNECTIONS;
   httpServerSettings.connections = httpConnections;
   //Specify the server's root directory
   strcpy(httpServerSettings.rootDirectory, "/www/");
   //Set default home page
   strcpy(httpServerSettings.defaultDocument, "index.html");
   //Callback functions
   httpServerSettings.requestCallback = httpServerRequestCallback;
   httpServerSettings.uriNotFoundCallback = httpServerUriNotFoundCallback;

   //HTTP server initialization
   error = httpServerInit(&httpServerContext, &httpServerSettings);
   //Failed to initialize HTTP server?
   if(error)
   {
      //Debug message
      TRACE_ERROR("Failed to initialize HTTP server!\r\n");
   }

   //Start HTTP server
   error = httpServerStart(&httpServerContext);
   //Failed to start HTTP server?
   if(error)
   {
      //Debug message
      TRACE_ERROR("Failed to start HTTP server!\r\n");
   }

#if defined (IAP_SCENARIO_3)
   //Load IAP public signature key
   error = iapUserLoadKeys();
   if (error)
   {
	  //Debug message
      TRACE_ERROR("Failed to load IAP user keys!\r\n");
      return error;
   }
#endif

   //Create IAP mutex
   if(!osCreateMutex(&iapMutex))
   {
      //Debug message
      TRACE_ERROR("Failed to create IAP mutex!\r\n");
   }

   //Create a task to blink the LED
   taskId = osCreateTask("LED", ledTask, NULL, 200, OS_TASK_PRIORITY_NORMAL);
   //Failed to create the task?
   if(taskId == OS_INVALID_TASK_ID)
   {
      //Debug message
      TRACE_ERROR("Failed to create task!\r\n");
   }

   //Start the execution of tasks
   osStartKernel();

   //This function should never return
   return 0;
}

/**
 * @brief Freertos User task stack overflow callback
 * @param[in] xTask Task handler of the task having a stack overflow
 * @param[in] pcTaskName Task name of the task having a stack overflow
 **/

void vApplicationStackOverflowHook(TaskHandle_t xTask,
                                       char *pcTaskName)
{
   TRACE_ERROR("Stack overflow in \"%s\"!\r\n", pcTaskName);
   vTaskSuspendAll();
   while(1)
   {
   }
}


/**
 * @brief Freertos User malloc failed callback
 **/

void vApplicationMallocFailedHook(void)
{
   TRACE_ERROR("Malloc failed!\r\n");
   vTaskSuspendAll();
   while(1)
   {
   }
}
